//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/Instrument/InstrumentListing.cpp
//! @brief     Implements class InstrumentListing.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/Instrument/InstrumentListing.h"
#include "GUI/Model/Device/InstrumentLibrary.h"
#include "GUI/Model/Project/ProjectDocument.h"
#include "GUI/View/Instrument/InstrumentLibraryDialog.h"
#include "GUI/View/Instrument/InstrumentListModel.h"
#include "GUI/View/Tool/Globals.h"
#include <QAction>
#include <QMessageBox>
#include <QVBoxLayout>

InstrumentListing::InstrumentListing()
    : m_instrument_library(std::make_unique<InstrumentLibrary>())
    , m_model(new InstrumentListModel)
    , m_separator_action1(new QAction(this))
    , m_separator_action2(new QAction(this))
{
    m_instrument_library->load();

    setSizePolicy(QSizePolicy::Preferred, QSizePolicy::Expanding);
    setViewMode(QListView::IconMode);
    setIconSize(QSize(96, 84));
    setMovement(QListView::Static);
    setFixedWidth(iconSize().width() + 10);
    setSelectionMode(QAbstractItemView::SingleSelection);

    setModel(m_model);

    m_new_gisas_action = new QAction("New GISAS", this);
    m_new_gisas_action->setIcon(QIcon(":/images/shape-square-plus.svg"));
    m_new_gisas_action->setToolTip("Add new GISAS instrument with default settings");
    connect(m_new_gisas_action, &QAction::triggered, this, &InstrumentListing::onNewGisas);
    addAction(m_new_gisas_action);

    m_new_offspec_action = new QAction("New off-specular", this);
    m_new_offspec_action->setIcon(QIcon(":/images/shape-square-plus.svg"));
    m_new_offspec_action->setToolTip("Add new off-specular instrument with default settings");
    connect(m_new_offspec_action, &QAction::triggered, this, &InstrumentListing::onNewOffspec);
    addAction(m_new_offspec_action);

    m_new_specular_action = new QAction("New specular", this);
    m_new_specular_action->setIcon(QIcon(":/images/shape-square-plus.svg"));
    m_new_specular_action->setToolTip("Add new specular instrument with default settings");
    connect(m_new_specular_action, &QAction::triggered, this, &InstrumentListing::onNewSpecular);
    addAction(m_new_specular_action);

    m_new_depthprobe_action = new QAction("New depth probe", this);
    m_new_depthprobe_action->setIcon(QIcon(":/images/shape-square-plus.svg"));
    m_new_depthprobe_action->setToolTip("Add new depth probe instrument with default settings");
    connect(m_new_depthprobe_action, &QAction::triggered, this,
            &InstrumentListing::onNewDepthprobe);
    addAction(m_new_depthprobe_action);

    m_separator_action1->setSeparator(true);
    addAction(m_separator_action1);

    m_remove_action = new QAction("Remove", this);
    m_remove_action->setIcon(QIcon(":/images/delete.svg"));
    m_remove_action->setToolTip("Remove selected instrument");
    connect(m_remove_action, &QAction::triggered, this, &InstrumentListing::onRemove);
    addAction(m_remove_action);

    m_copy_action = new QAction("Copy", this);
    m_copy_action->setIcon(QIcon(":/images/content-copy.svg"));
    m_copy_action->setToolTip("Make a copy of the selected instrument");
    connect(m_copy_action, &QAction::triggered, this, &InstrumentListing::onCopy);
    addAction(m_copy_action);

    m_separator_action2->setSeparator(true);
    addAction(m_separator_action2);

    m_store_in_library_action = new QAction("Store in library", this);
    m_store_in_library_action->setIcon(QIcon(":/images/library.svg"));
    m_store_in_library_action->setToolTip("Store instrument in library");
    connect(m_store_in_library_action, &QAction::triggered, this,
            &InstrumentListing::onStoreInLibrary);
    addAction(m_store_in_library_action);

    m_load_from_library_action = new QAction("Choose from library", this);
    m_load_from_library_action->setIcon(QIcon(":/images/library.svg"));
    m_load_from_library_action->setToolTip("Load an instrument from the instrument library");
    connect(m_load_from_library_action, &QAction::triggered, this,
            &InstrumentListing::onLoadFromLibrary);
    addAction(m_load_from_library_action);

    setContextMenuPolicy(Qt::ActionsContextMenu);

    updateActions();
    restoreSelection();

    connect(selectionModel(), &QItemSelectionModel::selectionChanged, this,
            &InstrumentListing::onItemSelectionChanged);
}

InstrumentListing::~InstrumentListing()
{
    m_instrument_library->saveIfModified();
}

QSize InstrumentListing::sizeHint() const
{
    return {170, 400};
}

QSize InstrumentListing::minimumSizeHint() const
{
    return {96, 200};
}

QList<QAction*> InstrumentListing::toolbarActions() const
{
    return {m_new_gisas_action,
            m_new_offspec_action,
            m_new_specular_action,
            m_new_depthprobe_action,
            m_separator_action1,
            m_remove_action,
            m_copy_action,
            m_separator_action2,
            m_store_in_library_action,
            m_load_from_library_action};
}

InstrumentItem* InstrumentListing::currentInstrumentItem() const
{
    const QModelIndexList indexes = selectionModel()->selectedIndexes();
    if (!indexes.empty())
        return m_model->instrumentItemForIndex(indexes.front());
    return nullptr;
}

void InstrumentListing::onItemSelectionChanged()
{
    updateActions();

    QModelIndexList indexes = selectionModel()->selectedIndexes();
    if (!indexes.empty()) {
        QModelIndex current = indexes.front();
        gDoc->instrumentModel()->setSelectedIndex(current.row());
        emit instrumentSelected(m_model->instrumentItemForIndex(current));
    } else
        emit instrumentSelected(nullptr);
}

void InstrumentListing::onNewGisas()
{
    QModelIndex idx = m_model->addNewGISASInstrument();
    selectionModel()->select(idx, QItemSelectionModel::ClearAndSelect);
}

void InstrumentListing::onNewOffspec()
{
    QModelIndex idx = m_model->addNewOffspecInstrument();
    selectionModel()->select(idx, QItemSelectionModel::ClearAndSelect);
}

void InstrumentListing::onNewSpecular()
{
    QModelIndex idx = m_model->addNewSpecularInstrument();
    selectionModel()->select(idx, QItemSelectionModel::ClearAndSelect);
}

void InstrumentListing::onNewDepthprobe()
{
    QModelIndex idx = m_model->addNewDepthprobeInstrument();
    selectionModel()->select(idx, QItemSelectionModel::ClearAndSelect);
}

//! Removes currently selected instrument.
void InstrumentListing::onRemove()
{
    QModelIndexList indexes = selectionModel()->selectedIndexes();
    if (!indexes.empty()) {
        m_model->removeInstrument(indexes.front());

        ensureItemSelected();
    }
}

//! Makes a copy of the currently selected instrument.
void InstrumentListing::onCopy()
{
    QModelIndexList indexes = selectionModel()->selectedIndexes();
    if (!indexes.empty()) {
        QModelIndex idx = m_model->copyInstrument(indexes.front());
        selectionModel()->select(idx, QItemSelectionModel::ClearAndSelect);
    }
}

void InstrumentListing::onStoreInLibrary()
{
    if (!selectionModel()->hasSelection())
        return;

    QModelIndex idx = selectionModel()->selectedIndexes().front();
    InstrumentItem* instrument = m_model->instrumentItemForIndex(idx);

    InstrumentLibraryDialog dlg(GUI::Global::mainWindow, m_instrument_library.get());
    dlg.execAdd(*instrument);
}

void InstrumentListing::onLoadFromLibrary()
{
    if (m_instrument_library->isEmpty()) {
        QMessageBox::information(GUI::Global::mainWindow, "Select from library",
                                 "The library does not contain instruments so far.");
        return;
    }

    InstrumentLibraryDialog dlg(GUI::Global::mainWindow, m_instrument_library.get());

    auto* instrumentToCopy = dlg.execChoose();
    if (instrumentToCopy == nullptr)
        return;

    QModelIndex idx = m_model->copyInstrument(instrumentToCopy);
    selectionModel()->select(idx, QItemSelectionModel::ClearAndSelect);
}

void InstrumentListing::updateActions()
{
    bool enabled = selectionModel()->hasSelection();
    m_remove_action->setEnabled(enabled);
    m_copy_action->setEnabled(enabled);
    m_store_in_library_action->setEnabled(enabled);
}

void InstrumentListing::ensureItemSelected()
{
    if (!selectionModel()->hasSelection() && m_model->rowCount()) {
        QModelIndex last = m_model->index(m_model->rowCount() - 1, 0, QModelIndex());
        selectionModel()->select(last, QItemSelectionModel::ClearAndSelect);
    }
}

void InstrumentListing::restoreSelection()
{
    int lastUsed = gDoc->instrumentModel()->selectedIndex();
    if (lastUsed >= 0 && lastUsed < m_model->rowCount()) {
        QModelIndex lastUsedIndex = m_model->index(lastUsed, 0, QModelIndex());
        selectionModel()->select(lastUsedIndex, QItemSelectionModel::ClearAndSelect);
    } else
        ensureItemSelected();

    updateActions();
}
