//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/Instrument/InstrumentsTreeModel.cpp
//! @brief     Implements class InstrumentsTreeModel.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2021
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/Instrument/InstrumentsTreeModel.h"
#include "Base/Util/Assert.h"
#include "GUI/Model/Device/InstrumentItems.h"
#include "GUI/Model/Device/InstrumentModel.h"
#include <QApplication>
#include <QtGui>

namespace {

using IType = InstrumentsTreeModel::InstrumentType;

const auto types = {IType::Gisas, IType::Offspec, IType::Specular, IType::Depthprobe};

} // namespace


InstrumentsTreeModel::InstrumentsTreeModel(QObject* parent, InstrumentModel* model)
    : QAbstractItemModel(parent)
    , m_model(model)
    , m_names_are_editable(false)
    , m_enable_empty_headlines(true)
{
}

void InstrumentsTreeModel::setNewInstrument(InstrumentItem* addedInstrument)
{
    m_new_instrument = addedInstrument;
}

void InstrumentsTreeModel::enableEmptyHeadlines(bool b)
{
    if (b == m_enable_empty_headlines)
        return;
    beginResetModel();
    m_enable_empty_headlines = b;
    endResetModel();
}

void InstrumentsTreeModel::clear()
{
    beginResetModel();
    endResetModel();
}

QVector<InstrumentItem*> InstrumentsTreeModel::instrumentItemsOfType(InstrumentType type) const
{
    switch (type) {
    case Gisas:
        return m_model->instrumentItems([](const InstrumentItem* p) {
            return dynamic_cast<const GISASInstrumentItem*>(p) != nullptr;
        });
    case Offspec:
        return m_model->instrumentItems([](const InstrumentItem* p) {
            return dynamic_cast<const OffspecInstrumentItem*>(p) != nullptr;
        });
    case Specular:
        return m_model->instrumentItems([](const InstrumentItem* p) {
            return dynamic_cast<const SpecularInstrumentItem*>(p) != nullptr;
        });
    case Depthprobe:
        return m_model->instrumentItems([](const InstrumentItem* p) {
            return dynamic_cast<const DepthprobeInstrumentItem*>(p) != nullptr;
        });
    default:
        return {};
    }
}

QModelIndex InstrumentsTreeModel::indexOfHeadline(InstrumentType type) const
{
    int row = 0;
    for (auto t : ::types) {
        if (t == type)
            return createIndex(row, 0, nullptr);
        row++;
    }

    return {};
}

QModelIndex InstrumentsTreeModel::index(int row, int column, const QModelIndex& parent) const
{
    if (!hasIndex(row, column, parent))
        return {};

    if (!parent.isValid())
        return createIndex(row, column, nullptr);

    for (auto type : ::types)
        if (parent == indexOfHeadline(type))
            return createIndex(row, column, instrumentItemsOfType(type)[row]);

    return {};
}

QModelIndex InstrumentsTreeModel::parent(const QModelIndex& index) const
{
    if (!index.isValid())
        return {};

    if (index.internalPointer() == nullptr) // index is headline => no parent
        return {};

    auto* item = itemForIndex(index);
    for (auto type : ::types)
        if (instrumentItemsOfType(type).contains(item))
            return indexOfHeadline(type);

    return {};
}

int InstrumentsTreeModel::columnCount(const QModelIndex& /*parent*/) const
{
    return 1;
}

int InstrumentsTreeModel::rowCount(const QModelIndex& parent) const
{
    if (!parent.isValid())
        return ::types.size();

    // parent is a headline
    for (auto type : ::types)
        if (parent == indexOfHeadline(type))
            return instrumentItemsOfType(type).size();

    return 0;
}

Qt::ItemFlags InstrumentsTreeModel::flags(const QModelIndex& index) const
{
    if (isHeadline(index) || !index.isValid())
        return Qt::NoItemFlags;

    auto f = QAbstractItemModel::flags(index);
    f |= Qt::ItemIsSelectable | Qt::ItemIsEnabled | Qt::ItemIsDragEnabled;

    if (index.column() == 0 && m_names_are_editable) // col 0 contains name of the data entry
        f |= Qt::ItemIsEditable;

    return f;
}

bool InstrumentsTreeModel::setData(const QModelIndex& index, const QVariant& value, int role)
{
    if (!index.isValid())
        return false;

    if (role == Qt::EditRole && index.column() == 0) {
        itemForIndex(index)->setInstrumentName(value.toString());
        emit dataChanged(index, index);
        return true;
    }

    if (role == Qt::ToolTipRole && index.column() == 0) {
        itemForIndex(index)->setDescription(value.toString());
        emit dataChanged(index, index);
        return true;
    }

    return false;
}

InstrumentItem* InstrumentsTreeModel::itemForIndex(const QModelIndex& index) const
{
    if (!index.isValid())
        return nullptr;

    return reinterpret_cast<InstrumentItem*>(index.internalPointer());
}

QModelIndex InstrumentsTreeModel::indexForItem(InstrumentItem* item) const
{
    if (item == nullptr)
        return {};

    for (auto type : ::types)
        if (auto row = instrumentItemsOfType(type).indexOf(item); row >= 0)
            return createIndex(row, 0, item);

    return {};
}

void InstrumentsTreeModel::removeItem(InstrumentItem* item)
{
    QModelIndex index = indexForItem(item);
    if (!index.isValid())
        return;

    beginRemoveRows(index.parent(), index.row(), index.row());
    m_model->removeInstrument(item);
    endRemoveRows();
}

bool InstrumentsTreeModel::isHeadline(const QModelIndex& index) const
{
    if (!index.isValid())
        return false;

    return index.internalPointer() == nullptr;
}

InstrumentsTreeModel::InstrumentType InstrumentsTreeModel::instrumentType(InstrumentItem* item)
{
    if (item->is<GISASInstrumentItem>())
        return Gisas;

    if (item->is<OffspecInstrumentItem>())
        return Offspec;

    if (item->is<SpecularInstrumentItem>())
        return Specular;

    if (item->is<DepthprobeInstrumentItem>())
        return Depthprobe;

    ASSERT_NEVER;
}

QVariant InstrumentsTreeModel::data(const QModelIndex& index, int role) const
{
    if (isHeadline(index)) {
        QString title;
        if (index == indexOfHeadline(Gisas))
            title = "GISAS";
        else if (index == indexOfHeadline(Offspec))
            title = "Offspec";
        else if (index == indexOfHeadline(Specular))
            title = "Specular";
        else if (index == indexOfHeadline(Depthprobe))
            title = "Depthprobe";

        switch (role) {
        case Qt::DisplayRole:
            return title;

        case Qt::FontRole: {
            QFont f(QApplication::font());
            f.setPointSize(f.pointSize() * 1.5);
            f.setBold(true);
            return f;
        }

        case Qt::SizeHintRole: {
            QFont f(QApplication::font());
            f.setPointSize(f.pointSize() * 1.5);
            f.setBold(true);
            QSize s = QFontMetrics(f).boundingRect(title).size();
            return QSize(s.width() * 2, s.height() * 2);
        }

        case Qt::TextAlignmentRole:
            return QVariant(Qt::AlignLeft | Qt::AlignVCenter);

        case Qt::BackgroundRole:
            return qApp->palette().color(QPalette::Base);

        case Qt::ForegroundRole:
            return qApp->palette().color(QPalette::Text);
        }
        return {};
    }

    auto* const item = itemForIndex(index);

    if (role == Qt::DisplayRole) {
        auto descr = item->description();
        if (!descr.isEmpty()) {
            descr.prepend("<br><br>");
            // max 4 lines
            while (descr.count("\n") > 3) {
                descr.truncate(descr.lastIndexOf("\n"));
                descr += " [...]";
            }
            descr.replace("\n", "<br>");
        }
        return "<b>" + item->instrumentName() + "</b>" + descr;
    }

    if (role == Qt::DecorationRole && (item == m_new_instrument)) {
        if (role == Qt::DecorationRole)
            switch (instrumentType(item)) {
            case Gisas:
                return QIcon(":/images/gisas_instrument_new.svg");
            case Offspec:
                return QIcon(":/images/offspec_instrument_new.svg");
            case Specular:
                return QIcon(":/images/specular_instrument_new.svg");
            case Depthprobe:
                return QIcon(":/images/depth_instrument_new.svg");
            default:
                break;
            }
    }

    if (role == Qt::ToolTipRole)
        return item->description();

    if (role == Qt::DisplayRole)
        return item->instrumentName();

    if (role == Qt::EditRole)
        return item->instrumentName();

    if (role == Qt::TextAlignmentRole)
        return QVariant(Qt::AlignLeft | Qt::AlignTop);

    if (role == Qt::DecorationRole)
        switch (instrumentType(item)) {
        case Gisas:
            return QIcon(":/images/gisas_instrument.svg");
        case Offspec:
            return QIcon(":/images/offspec_instrument.svg");
        case Specular:
            return QIcon(":/images/specular_instrument.svg");
        case Depthprobe:
            return QIcon(":/images/depth_instrument.svg");
        default:
            ASSERT_NEVER;
        }

    return {};
}
