//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/Tuning/SliderSettingsWidget.cpp
//! @brief     Implements class SliderSettingsWidget.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/Tuning/SliderSettingsWidget.h"
#include "GUI/Model/Data/Data2DItem.h"
#include "GUI/Model/Job/JobItem.h"
#include "GUI/Model/Project/ProjectDocument.h"
#include <QLabel>
#include <QVBoxLayout>

SliderSettingsWidget::SliderSettingsWidget()
    : m_radio1(nullptr)
    , m_radio2(nullptr)
    , m_radio3(nullptr)
    , m_lockz_check_box(nullptr)
{
    // tuning selectors
    QString tooltip("Allows to tune sample parameters within +/- of given range \nwith the help of "
                    "the slider.");

    auto* label = new QLabel("Tuning:");
    label->setToolTip(tooltip);

    m_radio1 = new QRadioButton("10%");
    m_radio1->setAutoExclusive(true);
    m_radio1->setToolTip(tooltip);
    connect(m_radio1, &QRadioButton::clicked, this, &SliderSettingsWidget::rangeChanged);

    m_radio2 = new QRadioButton("100%");
    m_radio2->setChecked(true);
    m_radio2->setAutoExclusive(true);
    m_radio2->setToolTip(tooltip);
    connect(m_radio2, &QRadioButton::clicked, this, &SliderSettingsWidget::rangeChanged);

    m_radio3 = new QRadioButton("1000%");
    m_radio3->setAutoExclusive(true);
    m_radio3->setToolTip(tooltip);
    connect(m_radio3, &QRadioButton::clicked, this, &SliderSettingsWidget::rangeChanged);

    // Fix z-axis
    m_lockz_check_box = new QCheckBox("Lock-Z");
    m_lockz_check_box->setToolTip(
        "Preserve (min, max) range of intensity axis during parameter tuning.");
    connect(m_lockz_check_box, &QCheckBox::stateChanged, this,
            &SliderSettingsWidget::onLockZChanged);

    auto* hbox = new QHBoxLayout;

    hbox->addWidget(label);
    hbox->addWidget(m_radio1);
    hbox->addWidget(m_radio2);
    hbox->addWidget(m_radio3);
    hbox->addStretch(1);
    hbox->addWidget(m_lockz_check_box);

    setLayout(hbox);
}

void SliderSettingsWidget::setJobItem(JobItem* job_item)
{
    m_job_item = job_item;

    initSlider();
    initZlock();
}

void SliderSettingsWidget::rangeChanged()
{
    if (!m_job_item)
        return;

    if (m_radio1->isChecked())
        m_job_item->setSliderRange(10);
    else if (m_radio2->isChecked())
        m_job_item->setSliderRange(100);
    else if (m_radio3->isChecked())
        m_job_item->setSliderRange(1000);

    emit sliderRangeFactorChanged(m_job_item->sliderRange());
    gDoc->setModified();
}

void SliderSettingsWidget::onLockZChanged(int state)
{
    if (state == Qt::Unchecked)
        emit lockzChanged(false);
    else if (state == Qt::Checked)
        emit lockzChanged(true);
    gDoc->setModified();
}

void SliderSettingsWidget::initSlider()
{
    if (!m_job_item)
        return;

    if (m_job_item->sliderRange() == 10)
        m_radio1->setChecked(true);
    else if (m_job_item->sliderRange() == 1000)
        m_radio3->setChecked(true);
    else {
        if (m_job_item->sliderRange() != 100)
            m_job_item->setSliderRange(100);
        m_radio2->setChecked(true);
    }
    emit sliderRangeFactorChanged(m_job_item->sliderRange());
}

void SliderSettingsWidget::initZlock()
{
    if (!m_job_item)
        return;

    if (auto* data_item = m_job_item->data2DItem())
        m_lockz_check_box->setChecked(data_item->isZaxisLocked());
}
